#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "mem.h"
#include "assert.h"
#include "command.h"
#include "rng.h"
#include "draw.h"
#include "user.h"
#include "log.h"

#ifndef NDEBUG
static int validCommand(const Command *c)
 {
  assert(c);
  
  if(c->data_start)
   {
    assert(c->data_read >= c->data_start);
    assert(c->data_read <= c->data_end);
    assert(c->data_start <= c->data_end);
    assert(c->data_end <= c->data_alloc);
    assert(c->data_start < c->data_alloc);
    assert(verifyPointer(c->data_start));
   }
  else
   {
    assert(!c->data_read);
    assert(!c->data_end);
    assert(!c->data_alloc);
   }
  
  return 1;
 }
#endif

void commandInit(Command *c, CommandType type)
 {
  assert(c);
  
  c->type = type;
  c->data_start = c->data_read = c->data_end = c->data_alloc = 0;
  
  assert(validCommand(c));
 }

int commandNetworkSend(const Command *c, Connection *con)
 {
  uint16_t size;
  uint8_t type, size_a, size_b;
  
  assert(validCommand(c));
  assert(con);
  
  size = c->data_end - c->data_start;
  type = c->type;
  size_a = size&255;
  size_b = size>>8;
  
  return networkSendBuffer(con, &type, &type+1) && networkSendBuffer(con, &size_a, &size_a+1) &&
         networkSendBuffer(con, &size_b, &size_b+1) && networkSendBuffer(con, c->data_start, c->data_end);
 }

int commandInitFromNetwork(Command *c, Connection *con)
 {
  uint16_t waiting;
  
  assert(c);
  assert(con);
  
  if((waiting = networkWaitingBytes(con)) > 2)
   {
    const uint8_t *data = networkGetReadBuffer(con);
    uint16_t size = data[1] | (data[2]<<8);
    
    if(waiting < size+3)
     return 0;
    
    c->type = data[0];
    
    if(size)
     {
      if(!(c->data_start = c->data_read = (uint8_t *)malloc(size)))
       return 0;
      
      c->type = data[0];
      c->data_alloc = c->data_end = c->data_start + size;
      memcpy(c->data_start, data+3, size);
     }
    else
     c->data_start = c->data_read = c->data_end = c->data_alloc = 0;
    
    networkSkipWaitingBytes(con, size+3);
    assert(validCommand(c));
    return 1;
   }
  
  return 0;
 }

void commandUninit(Command *c)
 {
  assert(validCommand(c));
  
  if(c->data_start)
   free(c->data_start);
 }

static int addCommandData(Command *c, uint8_t *data, uint32_t length)
 {
  assert(validCommand(c));
  
  if(length)
   {
    while((uint32_t)(c->data_alloc - c->data_end) <= length)
     {
      if(!c->data_start)
       {
        if(!(c->data_start = c->data_read = c->data_end = (uint8_t *)malloc(16)))
         return 0;
        
        c->data_alloc = c->data_start + 16;
       }
      else
       {
        uint8_t *old = c->data_start;
        
        if(!(c->data_start = (uint8_t *)realloc(old, 2*(c->data_alloc-c->data_start))))
         {
          c->data_start = old;
          return 0;
         }
        
        c->data_read = c->data_read - old + c->data_start;
        c->data_end = c->data_end - old + c->data_start;
        c->data_alloc = 2*(c->data_alloc - old) + c->data_start;
       }
     }
    
    memcpy(c->data_end, data, length);
    c->data_end += length;
   }
  
  assert(validCommand(c));
  return 1;
 }

/* Helpers for reading and writing data to commands: */

void commandResetReadPos(Command *c)
 {
  assert(validCommand(c));
  
  c->data_read = c->data_start;
 }

uint16_t commandGetReadPos(const Command *c)
 {
  assert(validCommand(c));
  
  return c->data_read - c->data_start;
 }

void commandSetReadPos(Command *c, uint16_t pos)
 {
  assert(validCommand(c));
  
  if((c->data_read = c->data_start + pos) > c->data_end)
   c->data_read = c->data_end;
  
  assert(validCommand(c));
 }

uint16_t commandRemainingBytes(Command *c)
 {
  assert(validCommand(c));
  
  return c->data_end - c->data_read;
 }

int commandPutU8(Command *c, uint8_t byte)
 {
  assert(validCommand(c));
  return addCommandData(c, &byte, 1);
 }

int commandPutU16(Command *c, uint16_t word)
 {
  assert(validCommand(c));
  return commandPutU8(c, word&255) && commandPutU8(c, (word>>8)&255);
 }

int commandPutU32(Command *c, uint32_t dword)
 {
  assert(validCommand(c));
  return commandPutU8(c, dword&255) && commandPutU8(c, (dword>>8)&255) && commandPutU8(c, (dword>>16)&255) && commandPutU8(c, dword>>24);
 }

int commandPutFloat(Command *c, double value)
 {
  assert(validCommand(c));
  
  if(value < 0) value = 0;
  else if(value > 400) value = 400;
  
  return commandPutU16(c, (uint32_t)(value/1000.0*0xffff));
 }

int commandPutString(Command *c, const char *string)
 {
  if(string) for(; *string; ++string)
   if(isprint(*string) && !commandPutS8(c, *string))
    return 0;
  
  return commandPutS8(c, 0);
 }

uint8_t commandGetU8(Command *c)
 {
  assert(validCommand(c));
  if(c->data_read != c->data_end)
  return *(c->data_read++);
  else return 0;
 }

uint16_t commandGetU16(Command *c)
 {
  assert(validCommand(c));
  return commandGetU8(c) | (commandGetU8(c)<<8);
 }

uint32_t commandGetU32(Command *c)
 {
  assert(validCommand(c));
  return commandGetU8(c) | (commandGetU8(c)<<8) | (commandGetU8(c)<<16) | (commandGetU8(c)<<24);
 }

double commandGetFloat(Command *c)
 {
  assert(validCommand(c));
  
  return (double)commandGetU16(c)/(double)0xffff*1000.0;
 }

void commandGetString(Command *c, char *string, uint16_t length)
 {
  char ch;
  
  assert(string);
  assert(length);
  
  while((ch = commandGetS8(c)))
   if(length > 1 && isprint(ch))
    {
     *string = ch;
     ++string;
     --length;
    }
  
  *string = 0;
 }

int commandHasOwner(const Command *c)
 {
  assert(validCommand(c));
  
  switch(c->type)
   {
    case Pen:
    case Pencil:
    case Brush:
    case Chalk:
    case Smudge:
     return 1;
    
    default:
     return 0;
   }
 }

uint8_t commandGetOwner(const Command *c)
 {
  assert(validCommand(c));
  
  if(commandHasOwner(c) && c->data_start != c->data_end)
   return *c->data_start;
  else
   return 0;
 }

void commandSetOwner(Command *c, uint8_t owner)
 {
  assert(validCommand(c));
  
  if(commandHasOwner(c) && c->data_start != c->data_end)
   *c->data_start = owner;
 }

void commandForgetOwner(Command *c)
 {
  assert(validCommand(c));
  
  commandSetOwner(c, 0);
 }

#ifdef CLIENT
#include "chat.h"

static int16_t *x = 0, *y = 0;
static int16_t points = 0, max_points = 0;

void commandPointsCleanup(void)
 {
  if(max_points)
   {
    assert(x && y);
    free(x);
    free(y);
    x = 0;
    y = 0;
    max_points = 0;
   }
  else
   {
    assert(!x && !y);
    
    /* actually we should just assert(0), how did we get here if we didn't allocate them? */
   }
 }

static int commandRemainderToPoints(Command *c)
 {
  if(!(points = commandRemainingBytes(c)/4))
   return 0;
  
  if(points > max_points)
   {
    if(!max_points)
     {
      assert(!x && !y);
      
      if(!(x = (int16_t*)malloc(points * sizeof *x)))
       return 0;
      
      if(!(y = (int16_t*)malloc(points * sizeof *y)))
       {
        free(x);
        x = 0;
        return 0;
       }
      
      atexit(&commandPointsCleanup);
     }
    else
     {
      int16_t *old_x = x, *old_y = y;
      
      assert(old_x && old_y);
      
      if(!(x = (int16_t*)realloc(old_x, points * sizeof *x)))
       {
        x = old_x;
        return 0;
       }
      
      if(!(y = (int16_t*)realloc(old_y, points * sizeof *y)))
       {
        y = old_y;
        return 0;
       }
     }
    
    max_points = points;
   }
  
  if(points)
   {
    int i = 0;
    for(; i < points; ++i)
     {
      assert(commandRemainingBytes(c));
      
      x[i] = commandGetS16(c);
      
      y[i] = commandGetS16(c);
     }
    
    return 1;
   }
  else return 0;
 }

extern uint8_t userid;

int commandPenInit(Command *c, Colour colour, const int16_t *x, const int16_t *y, uint32_t points)
 {
  uint32_t p = 0;
  
  assert(c);
  
  if(!points)
   return 0;
  
  assert(x && y);
  
  commandInit(c, Pen);
  
  if(!commandPutU8(c, userid) ||
     !commandPutU8(c, colour.r<0?0:colour.r>255?255:(uint8_t)colour.r) ||
     !commandPutU8(c, colour.g<0?0:colour.g>255?255:(uint8_t)colour.g) ||
     !commandPutU8(c, colour.b<0?0:colour.b>255?255:(uint8_t)colour.b))
  goto init_error;
  
  for(; p < points; ++p)
   if(!commandPutS16(c, x[p]) || !commandPutS16(c, y[p]))
    goto init_error;
  
  assert(validCommand(c));
  return 1;
  
  init_error:
  commandUninit(c);
  return 0;
 }

int commandPencilInit(Command *c, Colour colour, float radius, const int16_t *x, const int16_t *y, uint32_t points)
 {
  uint32_t p = 0;
  
  assert(c);
  
  if(!points)
   return 0;
  
  assert(x && y);
  
  commandInit(c, Pencil);
  
  if(!commandPutU8(c, userid) ||
     !commandPutU32(c, RNG()) ||
     !commandPutU8(c, colour.r<0?0:colour.r>255?255:(uint8_t)colour.r) ||
     !commandPutU8(c, colour.g<0?0:colour.g>255?255:(uint8_t)colour.g) ||
     !commandPutU8(c, colour.b<0?0:colour.b>255?255:(uint8_t)colour.b) ||
     !commandPutFloat(c, radius))
  goto init_error;
  
  for(; p < points; ++p)
   if(!commandPutS16(c, x[p]) || !commandPutS16(c, y[p]))
    goto init_error;
  
  assert(validCommand(c));
  return 1;
  
  init_error:
  commandUninit(c);
  return 0;
 }

int commandBrushInit(Command *c, Colour colour, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points)
 {
  uint32_t p = 0;
  
  assert(c);
  
  if(!points)
   return 0;
  assert(x && y);
  
  commandInit(c, Brush);
  
  if(!commandPutU8(c, userid) ||
     !commandPutU32(c, RNG()) ||
     !commandPutU8(c, colour.r<0?0:colour.r>255?255:(uint8_t)colour.r) ||
     !commandPutU8(c, colour.g<0?0:colour.g>255?255:(uint8_t)colour.g) ||
     !commandPutU8(c, colour.b<0?0:colour.b>255?255:(uint8_t)colour.b) ||
     !commandPutFloat(c, radius) ||
     !commandPutFloat(c, rate))
  goto init_error;
  
  for(; p < points; ++p)
   if(!commandPutS16(c, x[p]) || !commandPutS16(c, y[p]))
    goto init_error;
  
  assert(validCommand(c));
  return 1;
  
  init_error:
  commandUninit(c);
  return 0;
 }

int commandSmudgeInit(Command *c, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points)
 {
  uint32_t p = 0;
  
  assert(c);
  
  if(!points)
   return 0;
  
  assert(x && y);
  
  commandInit(c, Smudge);
  
  if(!commandPutU8(c, userid) ||
     !commandPutFloat(c, radius) ||
     !commandPutFloat(c, rate))
  goto init_error;
  
  for(; p < points; ++p)
   if(!commandPutS16(c, x[p]) || !commandPutS16(c, y[p]))
    goto init_error;
  
  assert(validCommand(c));
  return 1;
  
  init_error:
  commandUninit(c);
  return 0;
 }

int commandStampInit(Command *c, Colour colour, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points)
 {
  uint32_t p = 0;
  
  assert(c);
  
  if(!points)
   return 0;
  
  assert(x && y);
  
  commandInit(c, Stamp);
  
  if(!commandPutU8(c, userid) ||
     !commandPutU8(c, colour.r<0?0:colour.r>255?255:(uint8_t)colour.r) ||
     !commandPutU8(c, colour.g<0?0:colour.g>255?255:(uint8_t)colour.g) ||
     !commandPutU8(c, colour.b<0?0:colour.b>255?255:(uint8_t)colour.b) ||
     !commandPutFloat(c, radius) ||
     !commandPutFloat(c, rate))
  goto init_error;
  
  for(; p < points; ++p)
   if(!commandPutS16(c, x[p]) || !commandPutS16(c, y[p]))
    goto init_error;
  
  assert(validCommand(c));
  return 1;
  
  init_error:
  commandUninit(c);
  return 0;
 }

int commandChalkInit(Command *c, Colour colour, float radius, const int16_t *x, const int16_t *y, uint32_t points)
 {
  uint32_t p = 0;
  
  assert(c);
  
  if(!points)
   return 0;
  
  assert(x && y);
  
  commandInit(c, Chalk);
  
  if(!commandPutU8(c, userid) ||
     !commandPutU32(c, RNG()) ||
     !commandPutU8(c, colour.r<0?0:colour.r>255?255:(uint8_t)colour.r) ||
     !commandPutU8(c, colour.g<0?0:colour.g>255?255:(uint8_t)colour.g) ||
     !commandPutU8(c, colour.b<0?0:colour.b>255?255:(uint8_t)colour.b) ||
     !commandPutFloat(c, radius))
  goto init_error;
  
  for(; p < points; ++p)
   if(!commandPutS16(c, x[p]) || !commandPutS16(c, y[p]))
    goto init_error;
  
  assert(validCommand(c));
  return 1;
  
  init_error:
  commandUninit(c);
  return 0;
 }

void commandDraw(BITMAP *surface, const Command *c)
 {
  Command *_c = (Command *)c; /* We said it's const, but we're liars. (We'll set everything back when we're done, though.) */
  uint16_t readpos;
  
  enum
   {
    read_id = 0x0001,
    read_rng = 0x0002,
    read_colour = 0x0004,
    read_radius = 0x0008,
    read_rate = 0x0010,
    read_points = 0x0020
   };
  
  unsigned int read_flags = 0;
  
  uint8_t id = 0;
  Colour colour;
  float radius = 0.0;
  float rate = 0.0;
  
  assert(surface);
  assert(c);
  
  readpos = commandGetReadPos(_c);
  
  switch(c->type)
   {
    case Pen:
     read_flags = read_id | read_colour | read_points;
    break;
    
    case Pencil:
     read_flags = read_id | read_rng | read_colour | read_radius | read_points;
    break;
    
    case Brush:
     read_flags = read_id | read_rng | read_colour | read_radius | read_rate | read_points;
    break;
    
    case Smudge:
     read_flags = read_id | read_radius | read_rate | read_points;
    break;
    
    case Chalk:
     read_flags = read_id | read_rng | read_colour | read_radius | read_points;
    break;
    
    case Chat:
     read_flags = read_id;
    break;
    
    case Stamp:
      read_flags = read_id | read_colour | read_radius | read_rate | read_points;
    break;
    
    case Login:
     /* Login will ask to be drawn because it got added to the queue when we first
        logged in so we'd know when we were done loading. */
     goto done;
    
    case RepeatCommand:
     /* Should have been handled elsewhere since we have no idea what the command is
        we're supposed to repeat. Drop through to default for now. . . */
    
    default:
     logMessage(Warning, "Asked to draw invalid command (%d).", _c->type);
     assert(0);
     goto done;
   }
  
  if(read_flags & read_id)
   id = commandGetU8(_c);
  
  if(read_flags & read_rng)
   RNGPush(commandGetU32(_c));
  
  if(read_flags & read_colour)
   {
    colour.r = commandGetU8(_c);
    colour.g = commandGetU8(_c);
    colour.b = commandGetU8(_c);
    
    if(id) userSetColour(id, colour);
   }
  
  if(read_flags & read_radius)
   radius = commandGetFloat(_c);
  
  if(read_flags & read_rate)
   rate = commandGetFloat(_c);
  
  if(read_flags & read_points)
   {
    if(!commandRemainderToPoints(_c))
     {
      logMessage(Warning, "Invalid point data.");
      goto done;
     }
    
    if(id) userSetPos(id, *x, *y);
   }
  
  switch(c->type)
   {
    case Pen:
     drawPen(surface, colour, x, y, points);
    break;
    
    case Pencil:
     drawPencil(surface, colour, radius, x, y, points);
    break;
    
    case Brush:
     drawBrush(surface, colour, radius, rate, x, y, points);
    break;
    
    case Smudge:
     drawSmudge(surface, radius, rate, x, y, points);
    break;
    
    case Chalk:
     drawChalk(surface, colour, radius, x, y, points);
    break;
    
    case Stamp:
     drawStamp(surface, colour, radius, rate, x, y, points);
    break;
    
    case Chat:
     {
      char message[512];
      commandGetString(_c, message, sizeof message);
      chatAddMessage(id ? userGetName(id) : 0, message);
     }
    break;
    
    default:
     /* All cases should have been handled. */
     assert(0);
    break;
   }
  
  done:
  
  if(read_flags & read_rng)
   RNGPop();
  
  commandSetReadPos(_c, readpos);
 }

#endif

int commandInitFromFile(Command *c, bzf *file)
 {
  uint8_t type;
  
  assert(c);
  assert(file);
  
  if(!bz_Read(file, &type, 1))
   return 0;
  
  switch((c->type = type))
   {
    case Pen:
    case Brush:
    case Pencil:
    case Chalk:
    case Smudge:
    case Stamp:
     {
      uint8_t sizea, sizeb;
      uint16_t size;
      
      if(!bz_Read(file, &sizea, 1) ||
         !bz_Read(file, &sizeb, 1))
       return 0;
      
      size = (sizea<<8) | sizeb;
      
      if(!(c->data_start = c->data_read = (uint8_t*)malloc(size+1)))
       return 0;
      
      *c->data_start = 0;
      c->data_alloc = c->data_end = c->data_start + size + 1;
      
      if(!bz_Read(file, c->data_start+1, size))
       {
        free(c->data_start);
        return 0;
       }
      else
       return 1;
     }
    
    default:
     return 0;
   }
 }

int commandSave(Command *c, bzf *file)
 {
  assert(validCommand(c));
  assert(file);
   
  switch(c->type)
   {
    case Pen:
    case Brush:
    case Pencil:
    case Chalk:
    case Smudge:
    case Stamp:
     {
      uint16_t size;
      uint8_t sizea, sizeb, type;
      
      if(c->data_end - c->data_start < 1)
       return 0;
      
      size = c->data_end - c->data_start - 1;
      sizea = size>>8; 
      sizeb = size&255;
      type = c->type;
      
      return bz_Write(file, &type, 1) &&
             bz_Write(file, &sizea, 1) &&
             bz_Write(file, &sizeb, 1) &&
             bz_Write(file, c->data_start+1, size);
     }
    
    default:
     return 0;
   }
 }
