#ifndef COMMAND_H
#define COMMAND_H

#include <stdint.h>
#include "network.h"
#include "bzio.h"

typedef enum {Pen = 0,            /* Pen tool. */
              Pencil = 1,         /* Pencil tool. */
              Brush = 2,          /* Paint tool. */
              Chalk = 3,          /* Chalk tool. */
              Smudge = 4,         /* Smudge tool. */
              Chat = 5,           /* Chat command. */
              Login = 6,          /* Login. . . */
              Join = 7,           /* User joined. . . */
              Leave = 8,          /* User left. . . */
              Restart = 9,        /* Reset command. . . */
              Kick = 10,          /* Kick command. . . */
              Motd = 11,          /* Message of the day. . . */
              Stamp = 12,         /* Something like that. */
              
              RepeatCommand = 255 /* Server sends this for your commands rather than
                                     wasting bandwidth sending your commands back to you. */
             } CommandType; /* These have to be converted to a byte, so no more than 256 commands. */

typedef struct
 {
  CommandType type;
  uint8_t *data_start, *data_read, *data_end, *data_alloc;
 } Command;

extern void commandInit(Command *c, CommandType type);
extern void commandUninit(Command *c);

extern void commandResetReadPos(Command *c);
extern uint16_t commandGetReadPos(const Command *c);
extern uint16_t commandRemainingBytes(Command *c);
extern void commandSetReadPos(Command *c, uint16_t pos);

extern int commandPutU8(Command *c, uint8_t byte);
extern int commandPutU16(Command *c, uint16_t word);
extern int commandPutU32(Command *c, uint32_t dword);
extern int commandPutFloat(Command *c, double value);
extern int commandPutString(Command *c, const char *string);
extern uint8_t commandGetU8(Command *c);
extern uint16_t commandGetU16(Command *c);
extern uint32_t commandGetU32(Command *c);
extern double commandGetFloat(Command *c);
extern void commandGetString(Command *c, char *string, uint16_t maxlength);

extern uint8_t commandGetOwner(const Command *c);
extern void commandSetOwner(Command *c, uint8_t owner);
extern void commandForgetOwner(Command *c);

/* The signed versions of the above functions are really macros. */
#define commandPutS8(A, B)  commandPutU8(A,  (uint8_t)(B))
#define commandPutS16(A, B) commandPutU16(A, (uint16_t)(B))
#define commandPutS32(A, B) commandPutU32(A, (uint32_t)(B))
#define commandGetS8(A)     ((int8_t)commandGetU8(A))
#define commandGetS16(A)    ((int16_t)commandGetU16(A))
#define commandGetS32(A)    ((int32_t)commandGetU32(A))

extern int commandNetworkSend(const Command *c, Connection *con);
extern int commandInitFromNetwork(Command *c, Connection *con);

extern int commandInitFromFile(Command *c, bzf *file);
extern int commandSave(Command *c, bzf *file);

#ifdef CLIENT
    
    #include <allegro.h>
    #include "draw.h"
    
    extern int commandPenInit(Command *c, Colour colour, const int16_t *x, const int16_t *y, uint32_t points);
    extern int commandPencilInit(Command *c, Colour colour, float radius, const int16_t *x, const int16_t *y, uint32_t points);
    extern int commandBrushInit(Command *c, Colour colour, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points);
    extern int commandSmudgeInit(Command *c, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points);
    extern int commandChalkInit(Command *c, Colour colour, float radius, const int16_t *x, const int16_t *y, uint32_t points);
    extern int commandStampInit(Command *c, Colour colour, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points);
    
    extern void commandDraw(BITMAP *surface, const Command *c);
    
#endif

#endif
