#include <string.h>
#include <stdlib.h>

#include "mem.h"
#include "assert.h"
#include "commandqueue.h"

#ifndef NDEBUG
static int validCommandQueue(const CommandQueue *q)
 {
  assert(q);
  
  if(q->start)
   {
    assert(q->start <= q->end);
    assert(q->end <= q->alloc);
    assert(q->start < q->alloc);
   }
  else
   {
    assert(!q->end);
    assert(!q->alloc);
   }
  
  return 1;
 }
#endif

int commandQueueInit(CommandQueue *q)
 {
  assert(q);
  
  q->start = q->end = q->alloc = 0;
  
  assert(validCommandQueue(q));
  return 1;
 }

void commandQueueUninit(CommandQueue *q)
 {
  assert(validCommandQueue(q));
  
  if(q->start)
   {
    for(; q->end != q->start; --q->end)
     commandUninit(q->end-1);
    
    free(q->start);
   }
 }

int commandQueueForgetUser(CommandQueue *q, uint8_t user)
 {
  Command *c;
  
  assert(q);
  
  for(c = q->start; c != q->end; ++c)
   if(commandGetOwner(c) == user)
    commandForgetOwner(c);
  
  assert(validCommandQueue(q));
  return 1;
 }

static int commandQueueReserveExtraElement(CommandQueue *q)
 {
  assert(validCommandQueue(q));
  
  if(!q->start)
   {
    if(!(q->start = q->end = (Command *)malloc(sizeof(Command)*16)))
     return 0;
    
    q->alloc = q->start + 16;
    assert(validCommandQueue(q));
    return 1;
   }
  else
   {
    Command *old = q->start;
    
    if(!(q->start = (Command *)realloc(old, (q->alloc - q->start)*2*sizeof(Command))))
     {
      q->start = old;
      return 0;
     }
    
    q->end = q->end - old + q->start;
    q->alloc = (q->alloc - old)*2 + q->start;
    
    assert(validCommandQueue(q));
    return 1;
   }
 }

int commandQueuePushBack(CommandQueue *q, Command *c)
 {
  assert(validCommandQueue(q));
  assert(c);
  
  if(q->end == q->alloc && !commandQueueReserveExtraElement(q))
   return 0;
  
  *q->end = *c;
  ++q->end;
  
  return 1;
 }

int commandQueuePopBack(CommandQueue *q, Command *c)
 {
  assert(validCommandQueue(q));
  assert(c);
  
  if(q->end != q->start)
   {
    *c = *(--q->end);
    return 1;
   }
  else return 0;
 }

int commandQueuePushFront(CommandQueue *q, Command *c)
 {
  assert(validCommandQueue(q));
  assert(c);
  
  if(q->end == q->alloc && !commandQueueReserveExtraElement(q))
   return 0;
  
  memmove(q->start+1, q->start, sizeof(Command)*(q->end-q->start));
  *q->start = *c;
  ++q->end;
  
  return 1;
 }

int commandQueuePopFront(CommandQueue *q, Command *c)
 {
  assert(validCommandQueue(q));
  assert(c);
  
  if(q->end != q->start)
   {
    *c = *q->start;
    memmove(q->start, q->start+1, sizeof(Command)*(q->end-q->start-1));
    --q->end;
    return 1;
   }
  else return 0;
 }

uint16_t commandQueueElements(const CommandQueue *q)
 {
  assert(validCommandQueue(q));
  return q->end - q->start;
 }

const Command * commandQueueElement(const CommandQueue *q, uint16_t i)
 {
  assert(validCommandQueue(q));
  assert(i < q->end - q->start);
  
  return q->start+i;
 }

int commandQueueInitFromFile(CommandQueue *q, bzf *file)
 {
  Command c;
  
  assert(q);
  assert(file);
  
  if(commandQueueInit(q))
   {
    while(commandInitFromFile(&c, file))
     if(!commandQueuePushBack(q, &c))
      {
       commandUninit(&c);
       commandQueueUninit(q);
       return 0;
      }
    
    return 1;
   }
  
  return 0;
 }

int commandQueueSave(CommandQueue *q, bzf *file)
 {
  Command *c;
  
  assert(validCommandQueue(q));
  assert(file);
  
  for(c = q->start; c != q->end; ++c)
   if(!commandSave(c, file))
    return 0;
  
  return 1;
 }

int commandQueueInitFilename(CommandQueue *q, const char *filename)
 {
  bzf *bzfile = bz_OpenRead(filename);
  
  if(bzfile)
   {
    int result = commandQueueInitFromFile(q, bzfile);
    bz_CloseRead(bzfile);
    return result;
   }
  
  return 0;
 }

int commandQueueSaveFilename(CommandQueue *q, const char *filename)
 {
  bzf *bzfile = bz_OpenWrite(filename);
  
  if(bzfile)
   {
    int result = commandQueueSave(q, bzfile);
    bz_CloseWrite(bzfile);
    return result;
   }
  
  return 0;
 }
