#include "common.h"
#include "config.h"
#include "region.h"
#include "gui.h"

#include <string.h>

// Attributes a window can have:
const unsigned int attr_dirty  = 0x00000001; // Window contents has changed, and needs to be redrawn.
// const unsigned int attr_redraw = 0x00000002; // ...
const unsigned int attr_select = 0x00000004; // Window can be selected.
const unsigned int attr_move   = 0x00000008; // Window can moved.
const unsigned int attr_resize = 0x00000010; // Window can be resized.
const unsigned int attr_masked = 0x00000020; // Window is masked.

Window *root = 0; // This is the root window everything is drawn on.
static Window *selected = 0; // This is the window the user is interacting with.
static BITMAP *back_buffer = 0;
static DirtyRectangles list;

// --- Static functions that aren't used elsewhere. ---


static void GUIDrawSelectionBorder(Window *w)
 {
  assert(w);
  
  int v;
  
  acquire_bitmap(w->buffer);
  
  for(v = 0; v < w->buffer->w; ++v)
   {
    if((w->real_x+v+w->real_y)&1)
     _putpixel32(w->buffer, v, 0, ~_getpixel32(w->buffer, v, 0));
    
    if((w->real_x+w->buffer->w-1+w->real_y+v)&1)
     _putpixel32(w->buffer, v, w->buffer->h-1, ~_getpixel32(w->buffer, v, w->buffer->h-1));
   }
  
  for(v = 1; v < w->buffer->h-1; ++v)
   {
    if((w->real_x+w->real_y+v)&1)
     _putpixel32(w->buffer, 0, v,  ~_getpixel32(w->buffer, 0, v));
    
    if((w->real_x+w->buffer->w-1+w->real_y+v)&1)
     _putpixel32(w->buffer, w->buffer->w-1, v, ~_getpixel32(w->buffer, w->buffer->w-1, v));
   }
  
  release_bitmap(w->buffer);
  
  // TODO: When we can mark just part of a window dirty, mark just the outside instead of the whole thing.
  GUIMarkDirtyAll(w);
 }

static void GUIBringToFront(Window *w)
 {
  assert(w);
  
  if(w->parent)
   {
    // TODO: Do stuff here.
    
    GUIBringToFront(w->parent);
   }
 }

static void GUISelectWindow(Window *w)
 {
  assert(w);
  
  while(!(w->attr & attr_select)) // Window isn't selectable, we'll try it's parent.
   if(!(w = w->parent))
    return; // It doesn't have parent. Poor little window.
  
  if(selected != w)
   {
    if(selected)
     {
      // TODO: When we can mark just part of a window dirty, mark just the outside instead of the whole thing.
      GUIMarkDirtyAll(selected);
     }
    
    GUIDrawSelectionBorder(w);
    GUIBringToFront(w);
    selected = w;
   }
 }

static int GUIMouseWindow(Window *w, int x, int y, int xoff, int yoff, int b, int b_off, int decend)
 {
  assert(w);
  
  if(decend && w->start)
   {
    Window **child = w->start;
    for(; child != w->end; ++child)
     if(x >= (*child)->x && y >= (*child)->y && x < (*child)->x + (*child)->buffer->w && y < (*child)->y + (*child)->buffer->h)
      if(GUIMouseWindow(*child, x-(*child)->x, y-(*child)->y, xoff, yoff, b, b_off, 1))
       return 1;
   }
  
  if(w->mouse && (*w->mouse)(w, x, y, xoff, yoff, b, b_off))
   {
    GUISelectWindow(w);
    return 1;
   }
  
  return 0;
 }

static int GUIKeyWindow(Window *w, int value, int scancode, int decend)
 {
  if(decend && w->start)
   {
    Window **child = w->start;
    for(; child != w->end; ++child)
     if(GUIKeyWindow(*child, value, scancode, 1))
      return 1;
   }
  
  if(w->key && (*w->key)(w, value, scancode))
   {
    GUISelectWindow(w);
    return 1;
   }
  
  return 0;
 }

static void GUIClipObscured(DirtyRectangles * const dr, Window *const w, Window **child, int x, int y)
 {
  assert(w);
  assert(dr);
  
  if(DRIsDirty(dr))
   {
    Region rect = {x, y, x+w->buffer->w, y+w->buffer->h};
    DRClip(dr, &rect);
    
    for(; child != w->end; ++child)
     {
      rect.x1 = x+(*child)->x;
      rect.y1 = y+(*child)->y;
      rect.x2 = x+(*child)->x+(*child)->buffer->w;
      rect.y2 = y+(*child)->y+(*child)->buffer->h;
      
      DRRemove(dr, &rect);
     }
    
    if(w->parent)
     {
      child = w->parent->start;
      
      for(; child != w->parent->end; ++child)
       if(*child == w)
        {
         GUIClipObscured(dr, w->parent, child+1, x-w->x, y-w->y);
         return;
        }
      
      // Shouldn't have gotten here.
      assert(0);
     }
   }
 }

static void GUIRevealParent(Window * const w, const Region * const r)
 {
  assert(w);
  assert(r);
  
  if(w->parent)
   {
    if(w->parent->attr & attr_masked)
     {
      Region rect = {r->x1+w->x, r->y1+w->y, r->x2+w->x, r->y2+w->y};
      GUIRevealParent(w->parent, &rect);
      masked_blit(w->parent->buffer, back_buffer, w->x+r->x1, w->y+r->y1, w->real_x+r->x1, w->real_y+r->y1, r->x2-r->x1, r->y2-r->y1);
     }
    else
     blit(w->parent->buffer, back_buffer, w->x+r->x1, w->y+r->y1, w->real_x+r->x1, w->real_y+r->y1, r->x2-r->x1, r->y2-r->y1);
   }
 }

static int GUIUpdateWindow(Window *w)
 {
  assert(w);
  
  if(w->update && !(*(w->update))(w))
   {
    GUIDestroyWindow(w);
    return 0;
   }
  
  if(w->redraw && (w->attr & attr_dirty))
   {
    (*w->redraw)(w);
    GUIMarkDirtyAll(w);
    w->attr &= ~attr_dirty;
   }
  
  Window **child = w->start;
  for(; child != w->end; ++child)
   if(!GUIUpdateWindow(*child)) // TODO: If a child adds more children here, we're going to be in a lot of trouble. Protect against this later.
    --child;
  
  if(DRIsDirty(&w->dr)) // We might need to draw something. Oh no!
   {
    //if(selected == w)
    // GUIDrawSelectionBorder(w);
    
    
    GUIClipObscured(&w->dr, w, w->start, 0, 0);
    DROptimize(&w->dr);
    
    if(DRIsDirty(&w->dr)) // Yep, it's official. We need to draw something.
     {
      Region rect;
      
      if(DRIsSaturated(&w->dr))
       {
        assert(0); // This needs to be handled more carefully. It will draw on top of the children!
        rect.x1 = w->real_x;
        rect.y1 = w->real_y;
        rect.x2 = w->real_x+w->buffer->w;
        rect.y2 = w->real_y+w->buffer->h;
        
        DRAdd(&list, &rect);
        
        if(w->attr & attr_masked)
         {
          Region temp = {0, 0, w->buffer->w, w->buffer->h};
          GUIRevealParent(w, &temp);
          masked_blit(w->buffer, back_buffer, 0, 0, rect.x1, rect.y1, w->buffer->w, w->buffer->h);
         }
        else
         blit(w->buffer, back_buffer, 0, 0, rect.x1, rect.y1, w->buffer->w, w->buffer->h);
       }
      else
       {
        Region *r = w->dr.start;
        
        for(; r != w->dr.end; ++r)
         {
          rect.x1 = w->real_x+r->x1;
          rect.y1 = w->real_y+r->y1;
          rect.x2 = w->real_x+r->x2;
          rect.y2 = w->real_y+r->y2;
          
          assert(rect.x1 < rect.x2);
          assert(rect.y1 < rect.y2);
          
          DRAdd(&list, &rect);
          
          if(w->attr & attr_masked)
           {
            GUIRevealParent(w, r);
            masked_blit(w->buffer, back_buffer, r->x1, r->y1, rect.x1, rect.y1, rect.x2-rect.x1, rect.y2-rect.y1);
           }
          else
           blit(w->buffer, back_buffer, r->x1, r->y1, rect.x1, rect.y1, rect.x2-rect.x1, rect.y2-rect.y1);
          
          // TODO: Remove this
          // rect(screen, region.x1, region.y1, region.x2-1, region.y2-1, makecol(0, 0, 255));
         }
       }
      
      DRMakeClean(&w->dr);
     }
   }
  
  // TODO: Remove this.
  // rect(screen, w->real_x, w->real_y, w->real_x+w->buffer->w-1, w->real_y+w->buffer->h-1, makecol(0, 255, 0));
  
  return 1;
 }

static void GUIPropagateDirt(Window * const w, const Window * const stop, Region rect)
 {
  assert(w);
  //assert(stop);
  
  Region r = {0, 0, w->buffer->w, w->buffer->h};
  
  if(regionClip(&rect, &r))
   {
    DRAdd(&w->dr, &rect);
    
    Window **child = w->start;
    for(; child != w->end && *child != stop; ++child)
     {
      r.x1 = rect.x1-(*child)->x;
      r.y1 = rect.y1-(*child)->y;
      r.x2 = rect.x2-(*child)->x;
      r.y2 = rect.y2-(*child)->y;
      GUIPropagateDirt(*child, 0, r);
     }
   }
 }

static void GUIPropagateOffsets(Window *w, int xoff, int yoff)
 {
  assert(w);
  
  GUIMarkDirtyAll(w);
  
  w->real_x += xoff;
  w->real_y += yoff;
  
  if(w->start)
   {
    Window **child = w->start;
    
    for(; child != w->end; ++child)
     GUIPropagateOffsets(*child, xoff, yoff);
   }
 }

static int CreateRoot(Window *w, void *unused)
 {
  assert(w);
  
  BITMAP *pattern = load_bitmap("background.bmp", 0);
  
  if(pattern)
   { // Tile the pattern into the window.
    int x, y = 0;
    
    for(; y < w->buffer->h; y += pattern->h)
     for(x = 0; x < w->buffer->w; x += pattern->w)
      blit(pattern, w->buffer, 0, 0, x, y, pattern->w, pattern->h);
    
    destroy_bitmap(pattern);
   }
  else clear_to_color(w->buffer, makecol(100, 80, 70)); // Just make the window some default colour.
  
  // Since we've just drawn everything, and the root window shouldn't need to ever be redrawn, we'll clear the dirty flag.
  w->attr &= ~attr_dirty;
  
  return 1;
 }

// --- End of static functions that aren't used elsewhere. ---

static unsigned int init_count = 0;

int GUIInit(int width, int height, int depth, int fullscreen)
 {
  if(++init_count==1)
   {
    assert(!root);
    assert(!selected);
    assert(!back_buffer);
    
    DRInit(&list);
    
    set_color_depth(depth);
    
    if(set_gfx_mode(fullscreen ? GFX_AUTODETECT : GFX_AUTODETECT_WINDOWED, width, height, 0, 0))
     {
      GUIUninit();
      return 0;
     }
    
    show_mouse(screen);
    
    if(!(back_buffer = create_bitmap_ex(depth, width, height)))
     {
      GUIUninit();
      return 0;
     }
    
    if(!(root = GUICreateWindow(&CreateRoot, 0, 0, 0, width, height)))
     {
      GUIUninit();
      return 0;
     }
    
    selected = 0;
   }
  
  return 1;
 }

void GUIUninit(void)
 {
  if(!--init_count)
   {
    if(root)
     {
      GUIDestroyWindow(root);
      selected = 0;
      root = 0;
     }
    
    if(back_buffer)
     {
      destroy_bitmap(back_buffer);
      back_buffer = 0;
     }
    
    show_mouse(0);
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
   }
 }

void GUIUpdate(void)
 {
  assert(root && back_buffer && init_count);
  
  poll_mouse();
  poll_keyboard();
  
  static int old_b = 0, old_x = 0, old_y = 0;
  
  if(mouse_x != old_x || mouse_y != old_y || old_b != mouse_b)
   {
    int do_root = 1;
    if(selected && ((old_b || mouse_b) || (mouse_x >= selected->real_x && mouse_y >= selected->real_y &&
       mouse_x < selected->real_x + selected->buffer->w && mouse_y < selected->real_y + selected->buffer->h)))
     {
      if(GUIMouseWindow(selected, mouse_x-selected->real_x, mouse_y-selected->real_y, mouse_x-old_x, mouse_y-old_y, mouse_b, mouse_b ^ old_b, 1))
       do_root = 0;
     }
    
    if(do_root)
     GUIMouseWindow(root, mouse_x, mouse_y, mouse_x - old_x, mouse_y - old_y, mouse_b, mouse_b ^ old_b, 1);
   }
  
  while(keypressed())
   {
    int scancode, value = ureadkey(&scancode);
    
    if(!selected || !GUIKeyWindow(selected, value, scancode, 0))
     GUIKeyWindow(root, value, scancode, 1);
   }
  
  GUIUpdateWindow(root);
  
  if(DRIsDirty(&list))
   {
    if(DRIsSaturated(&list))
     {
      acquire_screen();
      
      if(mouse_sprite)
       {
        scare_mouse();
        blit(back_buffer, screen, 0, 0, 0, 0, back_buffer->w, back_buffer->h);
        unscare_mouse();
       }
      else
       blit(back_buffer, screen, 0, 0, 0, 0, back_buffer->w, back_buffer->h);
     }
    else
     {
      Region sr = {0, 0, back_buffer->w, back_buffer->h};
      DRClip(&list, &sr);
      DROptimize(&list);
      Region *r;
      
      #ifndef NDEBUG // Debugging - show outlines for the dirty rectangles.
       acquire_screen();
       scare_mouse();
       
       blit(back_buffer, screen, 0, 0, 0, 0, back_buffer->w, back_buffer->h);
       
       for(r = list.start; r != list.end; ++r)
        rect(screen, r->x1+1, r->y1+1, r->x2-2, r->y2-2, makecol(0, 0, 0));
       
       for(r = list.start; r != list.end; ++r)
        rect(screen, r->x1, r->y1, r->x2-1, r->y2-1, makecol(255, 255, 255));
       
       unscare_mouse();
      #else // Draw everything normally.
       if(mouse_sprite) // Is there a mouse on the screen?
        {
         freeze_mouse_flag = 1; // Prevent the mouse from moving while we're drawing.
         poll_mouse(); // Now that it's frozen, poll it to get the position.
         int mouse_shown = 1;
         Region mr = {mouse_x-mouse_x_focus, // This is the rectangle the mouse is living in.
                      mouse_y-mouse_y_focus,
                      mouse_x-mouse_x_focus+mouse_sprite->w,
                      mouse_y-mouse_y_focus+mouse_sprite->h};
         
         acquire_screen();
         
         for(r = list.start; r != list.end; ++r)
          {
           if(mouse_shown && regionCollision(r, &mr))
            { // The mouse is in the way of what we're drawing; hide it.
             scare_mouse();
             mouse_shown = 0;
            }
           
           blit(back_buffer, screen, r->x1, r->y1, r->x1, r->y1, r->x2-r->x1, r->y2-r->y1);
          }
         
         if(!mouse_shown)
          unscare_mouse();
         
         freeze_mouse_flag = 0;
        }
       else
        {
         acquire_screen();
         
         for(r = list.start; r != list.end; ++r)
          blit(back_buffer, screen, r->x1, r->y1, r->x1, r->y1, r->x2-r->x1, r->y2-r->y1);
        }
      #endif
     }
    
    release_screen();
    DRMakeClean(&list);
   }
  
  old_b = mouse_b;
  old_x = mouse_x;
  old_y = mouse_y;
 }

Window * GUICreateWindow(int (*init)(Window *, void *), void *init_data, int x, int y, int width, int height)
 {
  assert(width >= 1 && height >= 1);
  assert(back_buffer);
  
  Window *w = malloc(sizeof(Window));
  
  if(!w) return 0;
  
  w->attr = 0;
  w->update = 0;
  w->mouse = 0;
  w->redraw = 0;
  w->key = 0;
  w->destroy = 0;
  w->data = 0;
  w->x = w->real_x = x;
  w->y = w->real_y = y;
  w->parent = 0;
  w->start = w->end = w->alloc = 0;
  
  DRInit(&w->dr);
  
  if(!(w->buffer = create_bitmap_ex(bitmap_color_depth(back_buffer), width, height)))
   {
    free(w);
    return 0;
   }
  
  GUIMarkDirtyAll(w);
  
  #ifndef NDEBUG
  clear_to_color(w->buffer, makecol32(128, 128, 255));
  ellipse(w->buffer, width/2, height/2, width/2, height/2, makecol32(255, 255, 255));
  rect(w->buffer, 0, 0, width-1, height-1, makecol32(255, 255, 255));
  line(w->buffer, 0, 0, width-1, height-1, makecol32(255, 255, 255));
  line(w->buffer, width-1, 0, 0, height-1, makecol32(255, 255, 255));
  #endif
  
  if(init && !(*init)(w, init_data))
   {
    // In case the destroy function was set before the windows initialization failed.
    // Since it failed we shouldn't invoke it's destroy function.
    w->destroy = 0;
    
    GUIDestroyWindow(w);
    return 0;
   }
  
  if(w->redraw)
   (*w->redraw)(w);
  
  return w;
 }

void GUIDestroyWindow(Window *w)
 {
  assert(w);
  
  if(w == selected)
   selected = 0;
  
  if(w->destroy) // Let it know it's about to die a very painful death.
   (*(w->destroy))(w);
  
  if(w->start) // Is there a list of children?
   {
    while(w->start != w->end) // Destroy them all.
     {
      assert(w->start);
      --(w->end);
      assert((*w->end)->parent == w);
      (*w->end)->parent = 0;
      GUIDestroyWindow(*w->end);
     }
    
    free(w->start);
   }
  
  if(w->parent)
   GUIRemoveChild(w->parent, w);
  
  if(w->data)
   free(w->data);
  
  assert(w->buffer);
  destroy_bitmap(w->buffer);
  
  free(w);
 }

int GUIAddRoot(Window *w)
 {
  assert(w);
  
  return GUIAddChild(root, w);
 }

void GUIRemoveRoot(Window *w)
 {
  assert(w);
  
  GUIRemoveChild(root, w);
 }

int GUIAddChild(Window *w, Window *child)
 {
  assert(w);
  assert(child);
  assert(!child->parent);
  
  if(w->end == w->alloc)
   {
    if(!w->start)
     { // Allocate some memory.
      if(!(w->start = w->end = malloc(16 *sizeof *w->start)))
       return 0;
      
      w->alloc = w->start + 16;
     }
    else
     { // Resize the memory.
      Window **old = w->start;
      if(!(w->start = realloc(old, 2*(w->alloc-old)*sizeof *old)))
       {
        w->start = old;
        return 0;
       }
      
      w->end = w->end - old + w->start;
      w->alloc = (w->alloc - old)*2 + w->start;
     }
   }
  
  // child->real_x = w->real_x + child->x;
  // child->real_y = w->real_y + child->y;
  GUIPropagateOffsets(child, w->real_x + child->x - child->real_x, w->real_y + child->y - child->real_y);
  
  child->parent = w;
  GUIMarkDirtyAll(w);
  *(w->end) = child;
  ++(w->end);
  return 1;
 }

void GUIRemoveChild(Window *w, Window *child)
 {
  assert(w);
  assert(child);
  assert(child->parent == w);
  assert(w->start);
  
  Region rect = {child->x, child->y, child->x+child->buffer->w, child->y+child->buffer->h};
  GUIMarkDirty(w, &rect);
  
  Window **c = w->start;
  for(; c != w->end; ++c)
   if(*c == child)
    {
     memmove(c, c+1, (w->end-1-c)*sizeof*c);
     --(w->end);
     child->parent = 0;
     // child->real_x = child->x;
     // child->real_y = child->y;
     GUIPropagateOffsets(child, child->x - child->real_x, child->y - child->real_y); // Not sure this is right.
     return;
    }
  
  // Since window should have had the child, we shouldn't have gotten here.
  assert(0);
 }

void GUIMarkDirty(Window * const w, const Region * const r)
 {
  assert(w);
  assert(r);
  
  DRAdd(&w->dr, r);
 }

void GUIMarkDirtyGroup(Window * const w, const DirtyRectangles * const g)
 {
  assert(w);
  assert(g);
  
  DRAddGroup(&w->dr, 0, 0, g);
 }

void GUIMarkDirtyAll(Window * const w)
 {
  assert(w);
  
  Region rect = {0, 0, w->buffer->w, w->buffer->h};
  *w->dr.start = rect;
  w->dr.end = w->dr.start + 1;
 }

void GUIMoveWindow(Window *w, int xoff, int yoff)
 {
  assert(w);
  
  if(xoff || yoff)
   {
    if(w->parent)
     {
      Region rect = {w->x, w->y, w->x+w->buffer->w, w->y+w->buffer->h};
      GUIPropagateDirt(w->parent, w, rect);
     }
    
    GUIMarkDirtyAll(w);
    
    w->x += xoff;
    w->real_x += xoff;
    
    w->y += yoff;
    w->real_y += yoff;
    
    if(w->start)
     {
      Window **child = w->start;
      
      for(; child != w->end; ++child)
       GUIPropagateOffsets(*child, xoff, yoff);
     }
   }
 }

void GUIPositionWindow(Window *w, int x, int y)
 {
  assert(w);
  
  GUIMoveWindow(w, x-w->x, y-w->y);
 }

int GUIResizeWindow(Window *w, int width, int height)
 {
  assert(w);
  assert(width > 0 && height > 0);
  
  if((w->attr & attr_resize) && (width != w->buffer->w || height != w->buffer->h))
   {
    BITMAP *temp = create_bitmap_ex(32, width, height);
    
    if(!temp)
     return 0;
    
    if(w->parent && (width < w->buffer->w || height < w->buffer->h))
     {
      Region temp = {w->x, w->y, w->x+w->buffer->w, w->y+w->buffer->h};
      GUIMarkDirty(w->parent, &temp); // TODO: Mark only region that has changed.
     }
    
    destroy_bitmap(w->buffer);
    w->buffer = temp;
    
    // The window needs to be completely redrawn, now.
    w->attr |= attr_dirty;
    return 1;
   }
  
  return 0;
 }
