#ifndef REGION_H
#define REGION_H

#include <stdint.h>

typedef struct
 {
  int16_t x1, y1, x2, y2;
 } Region;

/* Returns non-zero if two regions overlap. Both regions must be normal. */
extern int regionCollision(const Region *a, const Region *b);

/* Returns non-zero if a region is normal. */
extern int regionIsNormal(const Region *r);

/* Normalize a region, making sure x1 <= x2 and y1 <= y2. */
extern void regionNormalize(Region *r);

/* Attempt to clip a region so that it fits inside another region. Both regions must be normal.
   Returns 0 on failure, in which case the regions state is left in an undefined state. */
extern int regionClip(Region *r, const Region *base);

typedef struct
 { /* We're going to be using a static buffer for the rectanges as
      opposed to allocating one automatically. If there isn't enough
      Room for all of them, end equals the end of the start array
      and the the entire area is considered dirty. */
  Region *end, start[256];
 } DirtyRectangles;

/* Initiate a DR structure. (They don't need to be uninitiated, however.) */
extern void DRInit(DirtyRectangles *dr);

/* Adds a region to the list. Region must be normal. Returns 0 on error, and list becomes saturated. */
extern int DRAdd(DirtyRectangles *dr, const Region *r);

/* Removes a region from the list. Region must be normal. Returns 0 on error, and the list becomes saturated. */
extern int DRRemove(DirtyRectangles *dr, const Region *r);

/* Add/Remove groups of rectanges from a region. */
extern int DRAddGroup(DirtyRectangles *dr, int16_t x, int16_t y, const DirtyRectangles *regions);
extern int DRRemoveGroup(DirtyRectangles *dr, int16_t x, int16_t y, const DirtyRectangles *regions);

/* Returns non-zero if the rectange list is saturated, in which case the entire region should be considered dirty. */
extern int DRIsSaturated(const DirtyRectangles *dr);
extern int DRIsDirty(const DirtyRectangles *dr);
extern int DRIsClean(const DirtyRectangles *dr);

/* Either make the entire list dirty, or make it clean. */
extern void DRMakeSaturated(DirtyRectangles *dr);
extern void DRMakeClean(DirtyRectangles *dr);

/* This function will reduce the number of rectangles. It's not perfect but it works fine in most cases. */
extern void DROptimize(DirtyRectangles * const dr);

/* Clips all the rectangles in the list to be within a region. */
extern void DRClip(DirtyRectangles *dr, const Region *r);

#endif
